//---------------------------------------------------------------------------
#ifndef UCalculatorH
#define UCalculatorH
//---------------------------------------------------------------------------
#include <string>
#include <vector>
#include <map>
#include <sstream>
#pragma hdrstop
using std::vector;
using std::string;
using std::map;
using std::pair;
using std::ostringstream;
//---------------------------------------------------------------------------

class TStringParser
{
  private:
    vector<string> FDictionary;    // Sorted with longest items first
    vector<string> FTokens;        // As located in the target string

    string FString;        // The string we are parsing
    bool   FTokenError;    // An unknown token was found (one not in the dictionary)
    bool   FCaseSensitive; // Obvious... but not yet implemented.

    void ParseTheTargetString();
    void FindToken(const string& TokenString, int Position, bool& TokenFound, string& Token);

  public:
    TStringParser() : FTokenError(false), FCaseSensitive(false) { ; }

    void InsertDictionaryItem(const string& AToken);
    void DeleteDictionaryItem(const string& AToken);
    bool LocateDictionaryItem(const string& AToken) const;

    void SetCaseSensitive(bool ACaseSensitive) { FCaseSensitive = ACaseSensitive; }
    void ProcessString(const string& AString)  { FString = AString; ParseTheTargetString(); }

    bool  GetCaseSensitive() const          { return FCaseSensitive; }
    bool  GetTokenError() const             { return FTokenError; }
    int   GetTokenCount() const             { return FTokens.size(); }
    const string& GetString() const         { return FString; }
    const string& GetToken(int Index) const { return FTokens[Index]; }
};
//---------------------------------------------------------------------------

#ifdef USE_LONG_DOUBLES
typedef long double TDouble;
#else
typedef double TDouble;
#endif
//---------------------------------------------------------------------------

#ifndef __STDC__
// Exception class used to promote any calls to matherr into C++ exceptions.
// What happens to matherr when __STDC__ is defined???
class EAStandardMathFunctionWentBang
{
#ifdef USE_LONG_DOUBLES
  private:
    _exceptionl* Exception;
  public:
    EAStandardMathFunctionWentBang(_exceptionl* E) : Exception(E) { ; }
#else
  private:
    _exception* Exception;
  public:
    EAStandardMathFunctionWentBang(_exception* E) : Exception(E) { ; }
#endif
};
#endif // __STDC__
//---------------------------------------------------------------------------

class TExpression;
//---------------------------------------------------------------------------

// Each token in an expression falls into one of the following broad categories.
// If you extend this enumeration it is vital to update the lookup table for
// valid sequences of tokens. Leave ttInvalidToken last since it governs the
// size of the lookup table.
enum TTokenType {ttOpenParenthesis, ttCloseParenthesis, ttVariable, ttFunction,
  ttNumber, ttConstant, ttUnary, ttBinary, ttInvalidToken};
//---------------------------------------------------------------------------

// Each token in an expression is one of the following types. The idea is to
// avoid having a separate function for every new token's factory.
const int stParenthesis = 0;

const int stPi          = 0;
const int stE           = 1;

const int stInteger     = 0;
const int stDecimal     = 1;
const int stExponent    = 2;

const int stSin         = 0;
const int stCos         = 1;
const int stTan         = 2;
const int stArcsin      = 3;
const int stArccos      = 4;
const int stArctan      = 5;

const int stSinh        = 0;
const int stTanh        = 1;
const int stCosh        = 2;
const int stArcsinh     = 3;
const int stArccosh     = 4;
const int stArctanh     = 5;

const int stAddition    = 0;
const int stSubtraction = 1;
const int stProduct     = 2;
const int stQuotient    = 3;
const int stPower       = 4;

const int stAbs         = 0;
const int stLn          = 1;
const int stLog         = 2;
const int stExp         = 3;
const int stMinus       = 4;
const int stSqrt        = 5;
//---------------------------------------------------------------------------

// These are the possible results when testing whether a given sequence of
// two tokens is allowed.
enum TTokenResult {trAllowed, trNotAllowed, trUnaryMinus, trImpliedProduct};
//---------------------------------------------------------------------------

// These are the possible errors values to be returned when an expression is
// invalid.
enum TTokenError {teNoError, teIllegalToken, teBadParentheses, teBadSequence,
  teMemoryAllocation, teCircularReference, teExceptionOccurred, teNoTokens,
  teInvalidNumber};
//---------------------------------------------------------------------------

// This function pointer type is used to implement arbitrary tokens in the
// expression evaluator - the client class implements the token in a
// derivative of TExpression and then supplies details about the token which
// can be used to create instances of it. The function is a static or global
// implemented by the client and returns a pointer to a TExpression which has
// been newly created.

// The reason for the This pointer is to allow specific objects to be referenced.
// This is important for TFunctions and TVariables.
typedef TExpression* (*TExpressionFactory)(void* This, int Subtype);
//---------------------------------------------------------------------------

// This simple structure contains everything we need to know about a token
// and how to use it.
struct TTokenInfo
{                                // Example:
  string             Token;      // "sin"   // External token format
  string             Private;    // "sin"   // Internal token format
  TTokenType         Type;       // ttUnary
  int                Subtype;    // stSin
  int                Priority;   // 6
  TExpressionFactory Factory;    // returns a pointer to a new TSinExpression
  void*              Object;     // Object on which to call Factory - this is
                                 // a late change intended to help move away from
                                 // using Borland's __closure extension.
};
//---------------------------------------------------------------------------

// This class interprets the output from a TStringParser to work out whether the
// expression is legal, and creates a tree of sub-expressions which easily
// evaluates the overall result.
class TExpressionEvaluator
{
  private:
    TStringParser            FParser;       // Parses the expression string to create a list of raw tokens.

    map<string, TTokenInfo*> FDictionary;   // Hold information about all the legal tokens.
    vector<string>           FTokens;       // Holds the normalised list of tokens after concatenating numbers.
    vector<TExpression*>     FExpressions;  // Holds a temporary list of pointers to expression objects.

    string FProcessedString;    // Holds the modified expression string.
    TExpression* FExpression;   // Holds a pointer to the final expression object after creating the tree.
    TTokenError  FTokenError;   // Holds the error value associated with the current string.

    // These are all called by ProcessString
    TTokenError NormaliseTheTokenList();
    TTokenError ValidateTheExpression();
    TTokenError CreateExpressionObjects();
    TTokenError EvaluateTheExpression();

    TTokenResult FTokenRelationships[ttInvalidToken + 1][ttInvalidToken + 1];
    TTokenResult IsLegalSequence(TTokenType FirstToken, TTokenType SecondToken) const;
    void SetupTokenRelationships();
    void SetupDictionaryItems();

    int GetHighestPriority(int Limit);

    // These factory functions are used to create the various types of
    // expression classes which are needed for the evaluator to proceed.
    static TExpression* GetConstantExpression(void* This, int Subtype);
    static TExpression* GetNumericExpression(void* This, int Subtype);
    static TExpression* GetTrigonometricExpression(void* This, int Subtype);
    static TExpression* GetHyperbolicExpression(void* This, int Subtype);
    static TExpression* GetBinaryExpression(void* This, int Subtype);
    static TExpression* GetOtherUnaryExpression(void* This, int Subtype);

    void InsertPrivateDictionaryItem(const string& Token, const string& Private, TTokenType Type,
      int Subtype, int Priority, TExpressionFactory Factory, void* Object);
    const TTokenInfo* LocatePrivateDictionaryItem(const string& AToken);

  public:
    TExpressionEvaluator();
    virtual ~TExpressionEvaluator();

    void InsertDictionaryItem(const string& Token, TTokenType Type, int Subtype, int Priority,
      TExpressionFactory Factory, void* Object);
    void DeleteDictionaryItem(const string& AToken);
    const TTokenInfo* LocateDictionaryItem(const string& AToken);

    TExpression* CreateExpressionObject(const string& AToken, int ParenthesisLevel);

    void ProcessString(const string& AString);
    const string& GetProcessedString() const { return FProcessedString; }
    TTokenError  GetTokenError() const       { return FTokenError; }
    TExpression* GetExpression() const       { return FExpression; }
};
//---------------------------------------------------------------------------

// This class knows how to use a token to operate on (up to) two operands. It
// uses the evaluator function for the token and passes it two operands. It
// is up to the client of TExpression to define the token and implement the
// function. The only exceptions are numeric constants, which are held in the
// FValue member (converted from the string token). This just seemed more
// efficient.
class TExpression
{
  friend class TExpressionEvaluator;
  protected:
    //const TExpressionEvaluator* FEvaluator; // Added to allow creation of partial derivatives.
    const TTokenInfo* FTokenInfo; // Information which determines what this expression does.
    TExpression*  FOperand1;      // Subexpression used as first operand.
    TExpression*  FOperand2;      // Subexpression used as second operand.
    string        FNumeric;       // Only used for stConstant, which are numbers.
    int           FParentheses;   // The parenthesis level of this expression.
    TDouble       FValue;         // Only used for stConstant.
    bool          FRemoved;       // Only used during intial evaluation of the token.

    string GetOperandString(const TExpression* Operand) const;
    string GetValueAsString() const
    {
      // Evaluate the expression.
      TDouble Result(0.0);
      try
      {
        // Actually retrieve the value.
        Result = GetValue();
      }
      catch(...)
      {
      }

      ostringstream ostr;
      ostr << Result;
      return ostr.str();
    }

  public:
    TExpression();
    virtual ~TExpression();

    void IncrementParentheses();
    void DecrementParentheses();

    // These were added to allow for the creation of partial derivatives.
    //TExpression* CreateDuplicateExpression();
    //virtual TExpression* CreateDerivativeExpression(const string& AToken) = 0;

    virtual TDouble GetValue() const = 0;

    const TTokenInfo* GetTokenInfo() const { return FTokenInfo; }
    int GetPriority() const;
    virtual bool GetIsConstant() const = 0;
    const string& GetNumeric();
    string GetRPNString() const;
    string GetNaturalString() const;

    virtual long GetNodeCount() const { return 1; }
    const TExpression* GetOperand1() const { return FOperand1; }
    const TExpression* GetOperand2() const { return FOperand2; }
    int GetParentheses() const { return FParentheses; }
};
//---------------------------------------------------------------------------

// The following classes are simple derivatives of TExpression which are
// intended to represent each of the operators supported by TExpressionEvaluator.
// They basically know how to evaluate a specific operator in terms of its
// operands. This is implemented through the virtual function GetValue.
// I have added another virtual which knows how to work out the expression for
// the partial derivative of the operator with respect to some (variable) token.

class TConstantExpression : public TExpression
{
  protected:
    int FConstantType;
    virtual TDouble GetValue() const;
    virtual bool GetIsConstant() const { return true; }
  public:
    TConstantExpression(int AConstantType) { FConstantType = AConstantType; }

    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TNumericExpression : public TExpression
{
  protected:
    virtual TDouble GetValue() const;
    virtual bool GetIsConstant() const { return true; }
  public:
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TUnaryExpression : public TExpression
{
  protected:
    virtual bool GetIsConstant() const { return FOperand1->GetIsConstant(); }
    virtual long GetNodeCount() const  { return FOperand1->GetNodeCount(); }
};
//---------------------------------------------------------------------------

class TBinaryExpression : public TExpression
{
  protected:
    virtual bool GetIsConstant() const { return FOperand1->GetIsConstant() && FOperand2->GetIsConstant(); }
    virtual long GetNodeCount() const  { return FOperand1->GetNodeCount() + FOperand2->GetNodeCount(); }
};
//---------------------------------------------------------------------------

class TAbsExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TSinExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TCosExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TTanExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcSinExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcCosExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcTanExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TLnExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TLogExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TExpExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TMinusExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TAdditionExpression : public TBinaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TSubtractionExpression : public TBinaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TProductExpression : public TBinaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TQuotientExpression : public TBinaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TPowerExpression : public TBinaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TSinhExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TCoshExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TTanhExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcSinhExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcCoshExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TArcTanhExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TSqrtExpression : public TUnaryExpression
{
  protected:
    virtual TDouble GetValue() const;
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

class TFunction;
class TVariable;
//---------------------------------------------------------------------------

// Class containing all the variables and functions.
class TCalculator
{
  private:
    // Variables encapsulate a double (or long double).
    map<string, TVariable*> FVariables;
    // Functions encapsulate an expression.
    map<string, TFunction*> FFunctions;
    // This does all the work of creating expressions.
    TExpressionEvaluator FExpressionEvaluator;

  public:
    TCalculator();
    virtual ~TCalculator();

    // Deleting a function or variable must re-parse all other functions.
    // I haven't done this yet - a minor task.

    void InsertVariable(const string& AToken);
    void DeleteVariable(const string& AToken);
    TVariable* GetVariable(const string& AToken);

    void InsertFunction(const string& AToken);
    void DeleteFunction(const string& AToken);
    TFunction* GetFunction(const string& AToken);

    TExpressionEvaluator& GetExpressionEvaluator() { return FExpressionEvaluator; }

    // Not yet implemented. All or most of the code which creates the derivatives is
    // already in place (and tested), but I have not exposed it yet.
    //TFunction* CreateDerivativeFunction(string Source, string Result,
    //  string Variable);
};
//---------------------------------------------------------------------------

// This simple class is used as a wrapper for the address of one of the variables.
// This needed so that TExpressionEvaluator can use the variable in the TExpression
// object it builds.
class TVariableExpression : public TExpression
{
  protected:
    const TDouble* FVariable;
    virtual TDouble GetValue() const    { return *FVariable; }
    virtual bool GetIsConstant() const  { return false; }

  public:
    explicit TVariableExpression(const TDouble* AVariable) : FVariable(AVariable) { ; }
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

// This is used to hold a function in TFunction. It's only real purpose is to
// help TFunction test for circular dependencies between the various functions.
// My solution seems simple enough, but I'm sure there is something cleverer.

// Here is my approach for finding circular dependencies:
// 1. Set a flag to indicate that the function is the is being called already.
// 2. If it is called again there is a circular reference and we can terminate the
//    evaluation. We return zero instead of recursing the function.
// 3. Set a flag to indicate circularity. This is static so that calling other
//    functions does not stop the mechanism working.
class TTopLevelExpression : public TExpression
{
  private:
    mutable bool FCalledAlready; // Only used internally so mutable OK.
    static bool FCircularError;
    static bool FExceptionOccurred;

  public:
    // Overrride the base class implementation to cope with the new flags.
    virtual TDouble GetValue() const;
    virtual bool GetIsConstant() const { return FOperand1->GetIsConstant(); }

  public:
    explicit TTopLevelExpression(TExpression* AExpression) { FOperand1 = AExpression; }
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);

    bool GetCircularError() const;
    bool GetExceptionOccurred() const { return FExceptionOccurred; }
    void ResetExceptionOccurred()     { FExceptionOccurred = false; }
};
//---------------------------------------------------------------------------

// This simple class is used as a wrapper for the address of one of the functions.
// This is needed so that TExpressionEvaluator can use the function in the TExpression
// object it builds.
class TFunctionExpression : public TExpression
{
  protected:
    TTopLevelExpression** FFunction;

    virtual TDouble GetValue() const;
    virtual bool GetIsConstant() const;

  public:
    explicit TFunctionExpression(TTopLevelExpression** AFunction) { FFunction = AFunction; }
    //virtual TExpression* CreateDerivativeExpression(const string& AToken);
};
//---------------------------------------------------------------------------

// Class for a variable - encapsulates a double (or long double).
class TVariable
{
  friend class TCalculator;
  private:
    TCalculator* FCalculator;  // The calculator which owns this variable.
    const string FToken;       // The variable has to have a unique name.

    TDouble FValue;

    static TExpression* GetVariableExpression(void* This, int Subtype)
      { return new TVariableExpression(&(static_cast<TVariable*>(This)->FValue)); }

  public:
    TVariable(const string& AToken, TCalculator* ACalculator) :
      FToken(AToken), FCalculator(ACalculator) { ; }

    const string& GetToken() const       { return FToken; }
    TDouble GetValue() const             { return FValue; }
    void SetValue(const TDouble& AValue) { FValue = AValue; }
};
//---------------------------------------------------------------------------

// This is a wrapper for a TFunctionExpression object which gives it a more friendly
// public interface. The life of the TFunction is not the same as for the
// TFunctionExpression, so inheritance was not considered.
class TFunction
{
  friend class TCalculator;
  private:
    TCalculator* FCalculator;       // The calculator which owns this function.
    const string FToken;            // The function has to have a unique name.
    TTokenError FTokenError;        // Holds the last value returned by SetString.
    string FNormalisedString;       // The expression string as normalised by the evaluator.

    TTopLevelExpression* FExpression;  // The current expression object.

    static TExpression* GetFunctionExpression(void* This, int Subtype)
      { return new TFunctionExpression(&(static_cast<TFunction*>(This)->FExpression)); }

  public:
    TFunction(const string& AToken, TCalculator* ACalculator) :
      FToken(AToken), FCalculator(ACalculator), FTokenError(teNoError),
      FExpression(0) { ; }
    virtual ~TFunction() { delete FExpression; }

    TTokenError SetString(const string& AExpressionString);

    bool GetCircular() const;
    bool GetException() const;
    TTokenError GetError() const       { return FTokenError; }
    const string& GetToken() const     { return FToken; }
    const string& GetString() const    { return FNormalisedString; }
    TDouble GetValue(); //const;
    string GetRPNString() const;
    string GetNaturalString() const;
};
//---------------------------------------------------------------------------

inline bool TFunction::GetCircular() const
{
  if (FExpression)
    return FExpression->GetCircularError();
  return false;
}
//---------------------------------------------------------------------------

inline bool TFunction::GetException() const
{
  if (FExpression)
    return FExpression->GetExceptionOccurred();
  return false;
}
//---------------------------------------------------------------------------

inline string TFunction::GetRPNString() const
{
  if (FExpression)
    return FExpression->GetOperand1()->GetRPNString();
  return "";
}
//---------------------------------------------------------------------------

inline string TFunction::GetNaturalString() const
{
  if (FExpression)
    return FExpression->GetOperand1()->GetNaturalString();
  return "";
}
//---------------------------------------------------------------------------

#endif
